#include "server.h"

inline unsigned char getPixelBlueScale(const SDL_Surface* restrict s, const int x, const int y)
{
	SDL_Color ret;
	Uint32 color = 0;
	char* position = (char*)s->pixels;

	position += (s->pitch * y);
	position += (s->format->BytesPerPixel * x);

	memcpy(&color, position, s->format->BytesPerPixel);

	SDL_GetRGB(color, s->format, &ret.r, &ret.g, &ret.b);

	return ret.b;
}

unsigned char** readMesh(const map* restrict mesh)
{
	/* Mesh data */
	unsigned char** restrict ret;

	/* Temporary SDL_Surface */
	SDL_Surface* restrict s;

	/* Loop variables */
	int i, j, k, l;

	/* Allocate memory for ret. */
	ret = (unsigned char**)scalloc(mesh->maptx * mesh->tilex, sizeof(unsigned char*));
	for(i = 0; i < mesh->maptx * mesh->tilex; i++)
	{
		ret[i] = (unsigned char*)scalloc(mesh->mapty * mesh->tiley, sizeof(unsigned char));
	}

	/* Read each file specified by mesh. */
	for(i = 0; i < mesh->maptx; i++)
	{
		for(j = 0; j < mesh->mapty; j++)
		{
			s = IMG_Load(mesh->tiles[i][j]);
			if(!s)
			{
				fprintf(stderr, "Could not locate mesh tile %s.  Aborting.\n", mesh->tiles[i][j]);
			}
			for(k = 0; k < mesh->tilex ; k++)
			{
				for(l = 0; l < mesh->tiley; l++)
				{
					ret[i*k][j*l] = getPixelBlueScale(s, k, l);
				}
			}
			SDL_FreeSurface(s);
		}
	}

	/* Basically a two-dimensional array of terrain types. */
	return ret;
}

int main(int argc __attribute__((unused)), char** argv __attribute__((unused)))
{
	/* The map to play on and the walk mesh to go with it. */
	map* m;
	map* mmesh;

	/* Filenames of the first map and mesh to load. */
	char* mapName;
	char* meshName;

	/* Storage for mesh data. */
	unsigned char** mesh;
	
	/* Array of clients */
	clientList clients;

	/* Array of avatars */
	avatarList avatars;

	/* Iteration var */
	unsigned short int i;

	/* Pointer array for passing done, listensock, clients and avatarsto thread funcs */
	void* dSAC[6];

	/* Pointer array for passing done, avatars and clients to serverphysics */
	void* dAAC[7];

	/* Thread for accepting new connections */
	SDL_Thread* acceptThread;

	/* Thread for physics computation and server->client communication */
	SDL_Thread* physicsThread;

	/* The log file we write to */
	char logfile[78];
	int log;

	/* The port to bind to */
	int port = -1;

	/* The maximum number of players */
	int players = 2;

	/* The socket listening on said port */
	TCPsocket listensock;

	/* False until we're done */
	bool done = false;

	/* Buffer for telling clients that the server is going down */
	char qBuf[sizeof(avatar)];

	/* Config file key-value pairs */
	kvlist* config;
	keyval* kv;

	/* Initialize mapName. */
	mapName = (char*)scalloc(80, sizeof(char));
	strcpy(mapName, MAPFILE);
	meshName = (char*)scalloc(80, sizeof(char));
	strcpy(meshName, MESHFILE);

	/* Print out a greeting. */
	fprintf(stdout, "SHAPEWARS: IT WILL BISECT YOUR FACE... SERVER CONSOLE\n");
	fprintf(stdout, "(press any key to quit)\n\n");

	/* Read the config file. */
	fprintf(stdout, "Reading configuration file...\n");
	config = readConfigFile(SERVERCONFIGNAME);
	while(!kvlistIsEmpty(config))
	{
		kv = kvlistPop(config);
		if(!strcmp(kv->key, "log"))
		{
			fprintf(stdout, "\tlog = %s\n", kv->val);
			strcpy(logfile, kv->val);
		}
		else if(!strcmp(kv->key, "port"))
		{
			fprintf(stdout, "\tport = %i\n", atoi(kv->val));
			port = atoi(kv->val);
		}
		else if(!strcmp(kv->key, "players"))
		{
			fprintf(stdout, "\tplayers = %i\n", atoi(kv->val));
			players = atoi(kv->val);
		}
		else if(!strcmp(kv->key, "map"))
		{
			fprintf(stdout, "\tmap = %s\n", kv->val);
			memset(mapName, 0, strlen(mapName));
			strcpy(mapName, kv->val);
		}
		else if(!strcmp(kv->key, "mesh"))
		{
			fprintf(stdout, "\tmesh = %s\n", kv->val);
			memset(meshName, 0, strlen(meshName));
			strcpy(meshName, kv->val);
		}

		sfree(kv->key);
		sfree(kv->val);
		kvalDestroy(kv);
	}
	kvlistDestroy(config);
	fprintf(stdout, "Done reading configuration file.\n");

	/* Set up server logging. */
	//log = open(logfile, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP);
	//dup2(log, 1);

	/* Load the map to play on and it's walk mesh. */
	fprintf(stdout, "Loading map and walk mesh...\n");
	m = loadMap(mapName);
	mmesh = loadMap(meshName);
	mesh = readMesh(mmesh);
	fprintf(stdout, "Done loading map and mesh.\n");
	fprintf(stdout, "Setting up socket and data structures.\n");
	if((listensock = makeServerSocket(port)) == NULL)
	{
		fprintf(stderr, "Server:\tCould not create server socket\n");
		return SOCKFAIL;
	}

	/* Initialize dSAC / dAAC */
	dSAC[0] = &done;
	dSAC[1] = &listensock;
	dSAC[2] = &clients;
	dSAC[3] = &avatars;
	dSAC[4] = &m->startx;
	dSAC[5] = &m->starty;

	dAAC[0] = &done;
	dAAC[1] = &avatars;
	dAAC[2] = &clients;
	dAAC[3] = &m->maptx;
	dAAC[4] = &m->tilex;
	dAAC[5] = &m->mapty;
	dAAC[6] = &m->tiley;

	/* We don't support larger servers */
	assert(players < 256);

	/* Allocate avatar and client arrays and initialize lists*/
	clients.cs = (client**)scalloc(players, sizeof(client*));
	avatars.as = (avatar**)scalloc(players, sizeof(avatar*));
	clients.max = players;
	clients.cur = 0;
	avatars.max = players;
	avatars.cur = 0;

	/* Initialize SDL for SDL_net */
	if(SDL_Init(0) < 0)
	{
		fprintf(stderr, "Could not SDL_Init()\n");
		return SDLINITFAIL;
	}

	if(SDLNet_Init() < 0)
	{
		fprintf(stderr, "Could not SDLNet_Init()\n");
		SDL_Quit();
		return SDLNETINITFAIL;
	}

	/* Initialize physics thread */
	physicsThread = SDL_CreateThread(serverphysics, dAAC);

	/* Start accepting connections from clients */
	acceptThread = SDL_CreateThread(serverAccept, dSAC);

	fprintf(stdout, "Ready to accept clients.\n");
	/* After any input, exit */ // TODO: Write a server console
	getchar();

	/* Signal to threads that we're done */
	done = true;

	/* Close our socket. */
	sclose(listensock);

	/* Kill and/or wait for threads. */
	SDL_KillThread(acceptThread); // TODO: why doesn't the below work?
	//SDL_WaitThread(acceptThread, NULL);
	SDL_WaitThread(physicsThread, NULL);

	/* Tell all client handling threads that we're done */
	for(i = 0; i < clients.max; i++)
	{
		if(clients.cs[i] != NULL)
		{
			qBuf[0] = 'Q';
			qBuf[1] = 'S';
			ssend(clients.cs[i]->sock, qBuf, sizeof(avatar));
			SDL_WaitThread(clients.cs[i]->thread, NULL);
			clients.cur--;
		}
	}

	/* Make sure everyone's quit. */
	assert(clients.cur == 0);

	/* Free stuff and exit. */
	sfree(clients.cs);
	sfree(avatars.as);

	freeMap(m);
	sfree(mapName);
	sfree(meshName);

	SDL_Quit();
	SDLNet_Quit();

	fprintf(stdout, "Server:\tAll threads are done, gracefully dying\n");

	return SUCCESS;
}
